@@ -5,6 +5,8 @@ import time
 
 from django.contrib.auth import get_user_model
 from djtezos.models import Blockchain, Contract, Call, Transfer, Transaction
+from djtezos.management.commands.djtezos_write import Command as Write
+from djtezos.management.commands.djtezos_balance import Command as Balance
 
 
 User = get_user_model()
@@ -48,10 +50,31 @@ def tzlocal():
     )
 
 
-@pytest.mark.django_db
-def test_story(user, tzlocal):
+@pytest.fixture
+def account(user, tzlocal):
     account = user.account_set.create(blockchain=tzlocal)
+    account.generate_private_key()
+    account.save()
+    return account
 
+
+def watch(tx, check):
+    tries = 100
+    while tries and not check(tx):
+        tx.sender.blockchain.provider.watch_blockchain(tx.sender.blockchain)
+        tx.refresh_from_db()
+        tries -= 1
+        time.sleep((100 - tries) / 2.0)
+    assert check(tx)
+
+
+def write():
+    Balance().handle()
+    Write().handle()
+
+
+@pytest.mark.django_db
+def test_story(user, account, tzlocal):
     contract = Transaction.objects.create(
         sender=account,
         contract_micheline=mich,
@@ -59,8 +82,8 @@ def test_story(user, tzlocal):
         args={'int': '1'},
         state='deploy',
     )
-    contract = Contract.objects.get(pk=contract.pk)
-    assert contract.state == 'done'
+    write()
+    watch(contract, lambda tx: tx.state == 'done')
 
     call = Transaction.objects.create(
         sender=account,
@@ -69,11 +92,13 @@ def test_story(user, tzlocal):
         args=[3],
         state='deploy',
     )
-    call = Call.objects.get(pk=call.pk)
-    assert call.state == 'done'
+    write()
+    watch(call, lambda tx: tx.state == 'done')
 
     balance = account.get_balance()
     account2 = user.account_set.create(blockchain=tzlocal)
+    account2.generate_private_key()
+    account2.save()
     balance2 = account2.get_balance()
     transfer = Transaction.objects.create(
         sender=account,
@@ -81,8 +106,9 @@ def test_story(user, tzlocal):
         amount=10000,
         state='deploy',
     )
-    transfer = Transfer.objects.get(pk=transfer.pk)
-    assert transfer.state == 'done'
+    write()
+    watch(transfer, lambda tx: tx.state == 'done')
+
     tries = 30
     while tries and not account.get_balance() < balance:
         time.sleep(1)
@@ -96,9 +122,7 @@ def test_story(user, tzlocal):
 
 
 @pytest.mark.django_db
-def test_wrong_storage(user, tzlocal):
-    account = user.account_set.create(blockchain=tzlocal)
-
+def test_wrong_storage(account):
     contract = Transaction.objects.create(
         sender=account,
         contract_micheline=mich,
@@ -106,15 +130,14 @@ def test_wrong_storage(user, tzlocal):
         args={'string': 'aoeu'},
         state='deploy',
     )
+    write()
     contract = Contract.objects.get(pk=contract.pk)
-    assert contract.state == 'deploy-aborted'
+    assert contract.state == 'retrying'
     assert contract.error
 
 
 @pytest.mark.django_db
-def test_wrong_args(user, tzlocal):
-    account = user.account_set.create(blockchain=tzlocal)
-
+def test_wrong_args(account):
     contract = Transaction.objects.create(
         sender=account,
         contract_micheline=mich,
@@ -122,7 +145,21 @@ def test_wrong_args(user, tzlocal):
         args={'int': '1'},
         state='deploy',
     )
-    contract = Contract.objects.get(pk=contract.pk)
+    write()
+    contract.refresh_from_db()
+    assert contract.state == 'watching'
+
+    # need now to call the watch function that will first either:
+    # - drop txhash and contract_address of all transactions of level above the
+    #   current head, to support the reorg case, this will allow users to retry
+    #   their transactions
+    # - or synchro new operations from current head level to last synchronized
+    #   level
+    # - and then poll the head for new level to start synchronizing on it
+    while not contract.contract_address:
+        account.blockchain.provider.watch_blockchain(account.blockchain)
+        contract.refresh_from_db()
+
     assert contract.state == 'done'
 
     call = Transaction.objects.create(
@@ -132,6 +169,7 @@ def test_wrong_args(user, tzlocal):
         args=['foobar'],
         state='deploy',
     )
+    write()
     call = Call.objects.get(pk=call.pk)
-    assert call.state == 'deploy-aborted'
+    assert call.state == 'retrying'
     assert call.error
